@@ -29,6 +29,13 @@ int56_div(value v1, value v2)
   CAMLreturn (copy_int56((Int64_val(v1) / divisor) << 8));
 }
 
+CAMLprim value
+int56_xor(value v1, value v2)
+{
+  CAMLparam2(v1, v2);
+  CAMLreturn (copy_int56((Int64_val(v1) ^ Int64_val(v2)) & mask));
+}
+
 CAMLprim value
 int56_shift_right(value v1, value v2)
 {
